import contextlib
import jinja2
import networkx as nx
import os
import yaml

from collections import defaultdict
from faker import Faker
from typing import *

from .factory_interface import FactoryInterface
from .types import Identifier
from .utils import normalize_identifiers, random_model, random_models


class FixturesLoader:
    """
    The factory "driver" class. Does most of the hard work of loading fixtures,
    leaving the responsibility of model instantiation up to the factory class
    passed in.
    """

    def __init__(self,
                 factory: FactoryInterface,
                 fixtures_dir: Optional[str] = None,
                 fixture_dirs: Optional[List[str]] = None,
                 env: Optional[jinja2.Environment] = None):
        """
        :param factory: An instance of the concrete factory to use for creating
                        models
        :param fixtures_dir: Path to folder to load template fixtures from
        :param env: An optional jinja environment (the default one will include
                    faker as a template global, but if you want to customize its
                    tags/filters/etc, then you need to create an env yourself - the
                    correct loader will be set automatically for you)
        """
        if not (fixtures_dir or fixture_dirs):
            raise TypeError('One of `fixtures_dir` or `fixture_dirs` is required.')

        self.env = self._ensure_env(env)
        """The Jinja Environment used for rendering the yaml template files."""

        factory.loader = self
        self.factory = factory
        """The factory instance."""

        self.fixture_dirs = fixture_dirs if fixture_dirs else [fixtures_dir]
        """A list of directories where fixture files should be loaded from."""

        self.relationships = {}
        """A dict keyed by model name where values are a list of related model names."""

        self.model_fixtures = defaultdict(dict)
        """A dict of models names to their semi-processed data from the yaml files."""

        self._cache = {}
        self._loaded = False

    def create_all(self, progress_callback: Optional[callable] = None) -> Dict[str, object]:
        """
        Creates all the models discovered from fixture files in :attr:`fixtures_dir`.

        :param progress_callback: An optional function to track progress. It must take three
                               parameters:
                                - an :class:`Identifier`
                                - the model instance
                                - and a boolean specifying whether the model was created
        :return: A dictionary keyed by identifier where the values are model instances.
        """
        if not self._loaded:
            self._load_data()

        dag = nx.DiGraph()
        for model_class_name, dependencies in self.relationships.items():
            dag.add_node(model_class_name)
            for dep in dependencies:
                dag.add_edge(model_class_name, dep)

        try:
            creation_order = reversed(list(nx.topological_sort(dag)))
        except nx.NetworkXUnfeasible:
            msg = 'Circular dependency detected between models'
            problem_graph = ', '.join([f'{a} -> {b}'
                                       for a, b in nx.find_cycle(dag)])
            raise Exception(f'{msg}: {problem_graph}')

        models = {}
        for model_class_name in creation_order:
            for identifier_key, data in self.model_fixtures[model_class_name].items():
                identifier = Identifier(model_class_name, identifier_key)
                data = self.factory.maybe_convert_values(identifier, data)
                self._cache[identifier_key] = data

                models[identifier_key], created = \
                    self.factory.create_or_update(identifier, data)
                if progress_callback:
                    progress_callback(identifier, models[identifier_key], created)

        self.factory.commit()
        return models

    def convert_identifiers(self, identifiers: Union[Identifier, List[Identifier]]):
        """
        Convert an individual :class:`Identifier` to a model instance,
        or a list of Identifiers to a list of model instances.
        """
        if not identifiers:
            return identifiers

        def _create_or_update(identifier):
            data = self._cache[identifier.key]
            return self.factory.create_or_update(identifier, data)[0]

        if isinstance(identifiers, Identifier):
            return _create_or_update(identifiers)
        elif isinstance(identifiers, list) and isinstance(identifiers[0], Identifier):
            return [_create_or_update(identifier) for identifier in identifiers]
        else:
            raise TypeError('`identifiers` must be an Identifier or list of Identifiers.')

    def _load_data(self):
        """
        Load all fixtures from :attr:`fixtures_dir`
        """
        filenames = []
        model_identifiers = defaultdict(list)
        for fixtures_dir in self.fixture_dirs:
            for filename in os.listdir(fixtures_dir):
                path = os.path.join(fixtures_dir, filename)
                file_ext = filename[filename.find('.')+1:]
                if os.path.isfile(path) and file_ext in {'yml', 'yaml'}:
                    filenames.append(filename)
                    with open(path) as f:
                        self._cache[filename] = f.read()

                    class_name = filename[:filename.rfind('.')]
                    with self._preloading_env() as env:
                        rendered_yaml = env.get_template(filename).render()
                        data = yaml.load(rendered_yaml)
                        model_identifiers[class_name] = list(data.keys())

        for filename in filenames:
            self._load_from_yaml(filename, model_identifiers)

        self._loaded = True

    def _load_from_yaml(self, filename: str, model_identifiers: Dict[str, List[str]]):
        """
        Load fixtures from the given filename
        """
        class_name = filename[:filename.rfind('.')]
        rendered_yaml = self.env.get_template(filename).render(
            model_identifiers=model_identifiers)
        fixture_data, self.relationships[class_name] = self._post_process_yaml_data(
            yaml.load(rendered_yaml),
            self.factory.get_relationships(class_name))

        for identifier_key, data in fixture_data.items():
            self.model_fixtures[class_name][identifier_key] = data

    def _post_process_yaml_data(self,
                                fixture_data: Dict[str, Dict[str, Any]],
                                relationship_columns: Set[str],
                                ) -> Tuple[Dict[str, Dict[str, Any]], List[str]]:
        """
        Convert and normalize identifier strings to Identifiers, as well as determine
        class relationships.
        """
        rv = {}
        relationships = set()
        for identifier_id, data in fixture_data.items():
            new_data = {}
            for col_name, value in data.items():
                if col_name not in relationship_columns:
                    new_data[col_name] = value
                    continue

                identifiers = normalize_identifiers(value)
                if identifiers:
                    relationships.add(identifiers[0].class_name)

                if isinstance(value, str) and len(identifiers) <= 1:
                    new_data[col_name] = identifiers[0] if identifiers else None
                else:
                    new_data[col_name] = identifiers

            rv[identifier_id] = new_data
        return rv, list(relationships)

    def _ensure_env(self, env: Union[jinja2.Environment, None]):
        """
        Make sure the jinja environment is minimally configured.
        """
        if not env:
            env = jinja2.Environment()
        if not env.loader:
            env.loader = jinja2.FunctionLoader(lambda filename: self._cache[filename])
        if 'faker' not in env.globals:
            faker = Faker()
            faker.seed(1234)
            env.globals['faker'] = faker
        if 'random_model' not in env.globals:
            env.globals['random_model'] = jinja2.contextfunction(random_model)
        if 'random_models' not in env.globals:
            env.globals['random_models'] = jinja2.contextfunction(random_models)
        return env

    @contextlib.contextmanager
    def _preloading_env(self):
        """
        A "stripped" jinja environment.
        """
        ctx = self.env.globals
        try:
            ctx['random_model'] = lambda *a, **kw: None
            ctx['random_models'] = lambda *a, **kw: None
            yield self.env
        finally:
            ctx['random_model'] = jinja2.contextfunction(random_model)
            ctx['random_models'] = jinja2.contextfunction(random_models)



import click
import importlib
import inspect
import os

from flask import current_app as app
from flask.cli import with_appcontext

try:
    from flask_sqlalchemy import Model as StockBaseModel
except ImportError:
    StockBaseModel = None
try:
    from flask_sqlalchemy_unchained import BaseModel as UnchainedBaseModel
except ImportError:
    UnchainedBaseModel = None

from ..fixtures_loader import FixturesLoader
from ..factories import SQLAlchemyModelFactory


@click.command()
@with_appcontext
def import_fixtures():
    models_module_name = app.config.get('PY_YAML_FIXTURES_MODELS_MODULE')
    try:
        models_module = importlib.import_module(models_module_name)
    except (ImportError, ModuleNotFoundError) as e:
        e.msg = (e.msg + '. Please make sure the PY_YAML_FIXTURES_MODELS_MODULE'
                         ' is set correctly.')
        raise e

    fixtures_dir = app.config.get('PY_YAML_FIXTURES_DIR')
    if not fixtures_dir or not os.path.exists(fixtures_dir):
        msg = ('Could not find the %r directory, please make sure '
               'PY_YAML_FIXTURES_DIR is set correctly' % fixtures_dir)
        raise NotADirectoryError(msg)

    model_classes = dict(inspect.getmembers(
        models_module,
        lambda obj: inspect.isclass(obj) and issubclass(obj, (
            StockBaseModel, UnchainedBaseModel))))

    factory = SQLAlchemyModelFactory(app.extensions['sqlalchemy'].db.session,
                                     model_classes)
    loader = FixturesLoader(factory, fixtures_dir=fixtures_dir)

    click.echo('Loading fixtures from %r for models in %r' % (
        fixtures_dir, models_module_name))
    loader.create_all(lambda identifier, model, created: click.echo(
        f'{"Creating" if created else "Updating"} {identifier.key}: {model!r}'))
    click.echo('Done adding fixtures')



from typing import *


class Identifier:
    def __init__(self, class_name: str, key: Union[int, str]):
        self.class_name = class_name
        self.key = key

    def __iter__(self):
        return iter([self.class_name, self.key])

    def __repr__(self):
        return '{cls}({key})'.format(cls=self.class_name, key=self.key)

    def __eq__(self, other):
        if not isinstance(other, Identifier):
            return False
        return self.class_name == other.class_name and self.key == other.key

    def __ne__(self, other):
        return not self.__eq__(other)



"""
    py_yaml_fixtures
    ~~~~~~~~~~~~~~~~

    A simple library to load database fixtures from Jinja-templated YAML files

    :copyright: Copyright © 2018 Brian Cappello
    :license: MIT, see LICENSE for more details
"""

__version__ = '0.3.2'


from .fixtures_loader import FixturesLoader



import datetime as dt
import functools

from sqlalchemy import orm as sa_orm
from sqlalchemy.ext.associationproxy import AssociationProxy
from types import FunctionType
from typing import *

from ..types import Identifier
from .. import utils
from ..factory_interface import FactoryInterface


class SQLAlchemyModelFactory(FactoryInterface):
    def __init__(self,
                 session: sa_orm.Session,
                 models: Union[List[type], Dict[str, type]],
                 date_factory: Optional[FunctionType] = None,
                 datetime_factory: Optional[FunctionType] = None):
        """
        :param session: the sqlalchemy session
        :param models: list of model classes, or dictionary of models by name
        :param date_factory: function used to generate dates (takes one
            parameter, the text value to convert)
        :param datetime_factory: function used to generate datetimes (takes one
            parameter, the text value to convert)
        """
        super().__init__()
        self.session = session
        self.models = (models if isinstance(models, dict)
                       else {model.__name__: model for model in models})
        self.model_instances = {}
        self.datetime_factory = datetime_factory or utils.datetime_factory
        self.date_factory = date_factory or utils.date_factory

    def create_or_update(self, identifier: Identifier, data: Dict[str, Any]):
        instance = self._get_existing(identifier, data)
        created = False
        if not instance:
            model_class = self.models[identifier.class_name]
            instance = model_class(**data)
            created = True
        else:
            for attr, value in data.items():
                setattr(instance, attr, value)

        self.session.add(instance)
        self.model_instances[identifier.key] = instance
        return instance, created

    def _get_existing(self, identifier: Identifier, data: Dict[str, Any]):
        model_class = self.models[identifier.class_name]
        instance = self.model_instances.get(identifier.key)
        if isinstance(instance, model_class) and instance in self.session:
            return instance

        # try to filter by primary key or any unique columns
        filter_kwargs = {}
        for col in model_class.__mapper__.columns:
            if col.name in data and (col.primary_key or col.unique):
                filter_kwargs[col.name] = data[col.name]

        # otherwise fallback to filtering by simple types
        if not filter_kwargs:
            filter_kwargs = {k: v for k, v in data.items()
                             if v is None
                             or isinstance(v, (bool, int, str, float))}
        if not filter_kwargs:
            return None

        with self.session.no_autoflush:
            return model_class.query.filter_by(**filter_kwargs).one_or_none()

    @functools.lru_cache()
    def get_relationships(self, class_name: str) -> Set[str]:
        rv = set()
        model_class = self.models[class_name]
        for col_name, value in model_class.__mapper__.all_orm_descriptors.items():
            if (isinstance(value, AssociationProxy)
                    or (hasattr(value, 'impl') and value.impl.uses_objects)):
                rv.add(col_name)
        return rv

    def maybe_convert_values(self,
                             identifier: Identifier,
                             data: Dict[str, Any],
                             ) -> Dict[str, Any]:
        model_class = self.models[identifier.class_name]
        relationships = self.get_relationships(identifier.class_name)
        rv = data.copy()
        for col_name, value in data.items():
            col = getattr(model_class, col_name)
            if col_name in relationships:
                rv[col_name] = self.loader.convert_identifiers(value)
            elif not hasattr(col, 'type'):
                continue
            elif col.type.python_type == dt.date:
                rv[col_name] = self.date_factory(value)
            elif col.type.python_type == dt.datetime:
                rv[col_name] = self.datetime_factory(value)
        return rv

    def commit(self):
        self.session.commit()



from typing import *

from .types import Identifier


class FactoryInterface:
    def __init__(self):
        from .fixtures_loader import FixturesLoader
        self.loader: FixturesLoader = None  # set by the FixturesLoader

    def create_or_update(self,
                         identifier: Identifier,
                         data: Dict[str, Any],
                         ) -> Tuple[object, bool]:
        """
        Create or update a model.

        :param identifier: An object with :attr:`class_name` and :attr:`key`
                           attributes
        :param data: A dictionary keyed by column name, with values being the
                     converted values to set on the model instance
        :return: A two-tuple of model instance and whether or not it was created.
        """
        raise NotImplementedError

    def get_relationships(self, class_name: str) -> Set[str]:
        """
        Return a list of model attribute names that could have relationships for
        the given model class name.

        :param class_name: The name of the class name to discover relationships for.
        :return: A set of model attribute names.
        """
        raise NotImplementedError

    def maybe_convert_values(self,
                             identifier: Identifier,
                             data: Dict[str, Any],
                             ) -> Dict[str, Any]:
        """
        Takes a dictionary of raw values for a specific identifier, as parsed
        from the YAML file, and depending upon the type of db column the data
        is meant for, decides what to do with the value (eg leave it alone,
        convert a string to a date/time instance, or convert identifiers to
        model instances by calling :meth:`self.loader.convert_identifiers`)

        :param identifier: An object with :attr:`class_name` and :attr:`key`
                           attributes
        :param data: A dictionary keyed by column name, with values being the
                     raw values as parsed from the YAML
        :return: A dictionary keyed by column name, with values being the
                 converted values meant to be set on the model instance
        """
        raise NotImplementedError

    def commit(self):
        """
        If your ORM implements the data mapper pattern instead of active
        record, then you can implement this method to commit the session after
        all the models have been added to it.
        """
        pass



import random
import re

from collections import defaultdict
from datetime import datetime, timezone
from dateutil.parser import parse as parse_datetime
from typing import *

from .types import Identifier


IDENTIFIER_RE = re.compile(r'(?P<class_name>\w+)\((?P<identifiers>[\w,\s]+)\)')


def datetime_factory(value):
    if value in {'today', 'now', 'utcnow'}:
        return datetime.now(timezone.utc)
    elif value not in {None, 'None'}:
        return parse_datetime(value)


def date_factory(value):
    dt = datetime_factory(value)
    if isinstance(dt, datetime):
        return dt.date


def random_model(ctx, model_class_name):
    model_identifiers = ctx['model_identifiers'][model_class_name]
    if not model_identifiers:
        return 'None'
    idx = random.randrange(0, len(model_identifiers))
    return '%s(%s)' % (model_class_name, model_identifiers[idx])


def random_models(ctx, model_class_name, min_count=0, max_count=3):
    model_identifiers = ctx['model_identifiers'][model_class_name]
    num_models = random.randint(min_count, min(max_count, len(model_identifiers)))
    if num_models == 0:
        return '[]'

    added = set()
    while len(added) < num_models:
        idx = random.randrange(0, len(model_identifiers))
        added.add(model_identifiers[idx])
    return '["%s(%s)"]' % (model_class_name, ','.join(added))

def normalize_identifiers(identifiers: Union[str, List[str]]) -> List[Identifier]:
    if not identifiers:
        return identifiers

    if isinstance(identifiers, str):
        identifiers = _convert_str(identifiers)
    if isinstance(identifiers, (list, tuple)):
        identifiers = _group_by_class_name(identifiers)

    rv = {}
    for class_name, values in identifiers.items():
        if not class_name:
            raise Exception('Identifier must have a class name.')
        for key in _flatten_csv_list(values):
            if not key:
                continue
            rv[key] = Identifier(class_name, key)
    return list(rv.values())


def _group_by_class_name(identifiers: List[str]) -> DefaultDict[str, List[str]]:
    rv = defaultdict(list)
    for v in identifiers:
        if isinstance(v, Identifier):
            rv[v.class_name].append(v.key)
        elif isinstance(v, str):
            for identifier in _convert_str(v):
                rv[identifier.class_name].append(identifier.key)
        else:
            raise Exception(
                'Unexpected type {t} (for {v!r})'.format(t=type(v), v=v))
    return rv


def _flatten_csv_list(identifier_keys: List[str]) -> List[str]:
    return [key.strip()
            for keys in identifier_keys
            for key in keys.strip(',').split(',')]


def _convert_str(value: str) -> List[Identifier]:
    value = ''.join(value.splitlines())
    rv = []
    prev = None
    while True:
        match = IDENTIFIER_RE.search(value, prev.end() if prev else 0)
        if not match and not rv:
            raise Exception(f'Identifier must have a class name. {value}')
        elif not match:
            return rv

        rv.append(Identifier(match.group('class_name'),
                             match.group('identifiers')))
        prev = match



from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='Py YAML Fixtures',
    version='0.3.2',
    description='Load database fixtures from in Jinja-templated YAML files',
    long_description=long_description,
    url='https://github.com/briancappello/py-yaml-fixtures',
    author='Brian Cappello',
    license='MIT',

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'faker>=0.8.7',
        'jinja2>=2.10',
        'networkx>=2.2',
        'python-dateutil>=2.6.1',
        'PyYAML>=3.12',
    ],
    extras_require={
        'dev': [
            'pytest',
            'sqlalchemy>=1.0',
        ],
        'docs': [
            'm2r',
            'sphinx',
            'sphinx-autobuild',
            'sphinx-rtd-theme',
        ],
        'sqlalchemy': [
            'sqlalchemy>=1.0',
        ],
    },
    python_requires='>=3.5',
    include_package_data=True,
    zip_safe=False,
)



from .cli import import_fixtures


class PyYAMLFixtures:
    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault('PY_YAML_FIXTURES_MODELS_MODULE', 'app.models')
        app.config.setdefault('PY_YAML_FIXTURES_DIR', 'db/fixtures')
        command_name = app.config.setdefault('PY_YAML_FIXTURES_COMMAND_NAME',
                                             'import-fixtures')

        if command_name:
            app.cli.add_command(import_fixtures, command_name)
